@@ -2,9 +2,13 @@
 
 from __future__ import absolute_import
 
+import os
+import glob
 import sys
 import re
 import threading
+import collections
+import itertools
 
 try:
     from collections.abc import Mapping
@@ -18,12 +22,17 @@ try:
     import queue
 except ImportError:
     import Queue as queue
+try:
+    from urllib.parse import urlparse
+except ImportError:
+    from urlparse import urlparse
 
 import numpy
 
+import uproot4.cache
 import uproot4.source.cursor
 import uproot4.streamers
-import uproot4.stl_containers
+import uproot4.containers
 import uproot4.interpretation
 import uproot4.interpretation.numerical
 import uproot4.interpretation.jagged
@@ -78,10 +87,14 @@ def _regularize_executors(decompression_executor, interpretation_executor):
 def _regularize_array_cache(array_cache, file):
     if isinstance(array_cache, MutableMapping):
         return array_cache
-    elif array_cache is None:
+    elif array_cache is None and file is not None:
         return file._array_cache
+    elif array_cache is None:
+        return None
+    elif uproot4._util.isint(array_cache) or uproot4._util.isstr(array_cache):
+        return uproot4.cache.LRUArrayCache(array_cache)
     else:
-        raise TypeError("array_cache must be None or a MutableMapping")
+        raise TypeError("array_cache must be None, a MutableMapping, or a memory size")
 
 
 def _regularize_aliases(hasbranches, aliases):
@@ -403,6 +416,7 @@ def _ranges_or_baskets_to_arrays(
         branchid_num_baskets[id(branch)] += 1
 
         if isinstance(range_or_basket, tuple) and len(range_or_basket) == 2:
+            range_or_basket = (int(range_or_basket[0]), int(range_or_basket[1]))
             ranges.append(range_or_basket)
             range_args[range_or_basket] = (branch, basket_num)
             range_original_index[range_or_basket] = original_index
@@ -498,6 +512,57 @@ def _ranges_or_baskets_to_arrays(
             raise AssertionError(obj)
 
 
+def _hasbranches_num_entries_for(
+    hasbranches, target_num_bytes, entry_start, entry_stop, branchid_interpretation
+):
+    total_bytes = 0.0
+    for branch in hasbranches.itervalues(recursive=True):
+        if id(branch) in branchid_interpretation:
+            entry_offsets = branch.entry_offsets
+            start = entry_offsets[0]
+            for basket_num, stop in enumerate(entry_offsets[1:]):
+                if entry_start < stop and start <= entry_stop:
+                    total_bytes += branch.basket_compressed_bytes(basket_num)
+                start = stop
+
+    total_entries = entry_stop - entry_start
+    num_entries = int(round(target_num_bytes * total_entries / total_bytes))
+    if num_entries <= 0:
+        return 1
+    else:
+        return num_entries
+
+
+def _regularize_step_size(
+    hasbranches, step_size, entry_start, entry_stop, branchid_interpretation
+):
+    if uproot4._util.isint(step_size):
+        return step_size
+    target_num_bytes = uproot4._util.memory_size(
+        step_size,
+        "number of entries or memory size string with units "
+        "(such as '100 MB') required, not {0}".format(repr(step_size)),
+    )
+    return _hasbranches_num_entries_for(
+        hasbranches, target_num_bytes, entry_start, entry_stop, branchid_interpretation
+    )
+
+
+Report = collections.namedtuple(
+    "Report",
+    [
+        "global_entry_start",
+        "global_entry_stop",
+        "tree_entry_start",
+        "tree_entry_stop",
+        "container",
+        "tree",
+        "file",
+        "file_path",
+    ],
+)
+
+
 class HasBranches(Mapping):
     @property
     def branches(self):
@@ -802,6 +867,26 @@ class HasBranches(Mapping):
         library="ak",
         how=None,
     ):
+        keys = set(self.keys(recursive=True, full_paths=False))
+        if isinstance(self, TBranch) and expressions is None and len(keys) == 0:
+            filter_branch = uproot4._util.regularize_filter(filter_branch)
+            return self.parent.arrays(
+                expressions=expressions,
+                cut=cut,
+                filter_name=filter_name,
+                filter_typename=filter_typename,
+                filter_branch=lambda branch: branch is self and filter_branch(branch),
+                aliases=aliases,
+                compute=compute,
+                entry_start=entry_start,
+                entry_stop=entry_stop,
+                decompression_executor=decompression_executor,
+                interpretation_executor=interpretation_executor,
+                array_cache=array_cache,
+                library=library,
+                how=how,
+            )
+
         entry_start, entry_stop = _regularize_entries_start_stop(
             self.tree.num_entries, entry_start, entry_stop
         )
@@ -825,7 +910,6 @@ class HasBranches(Mapping):
             else:
                 return None
 
-        keys = set(self.keys(recursive=True, full_paths=False))
         aliases = _regularize_aliases(self, aliases)
         arrays, expression_context, branchid_interpretation = _regularize_expressions(
             self,
@@ -892,6 +976,193 @@ class HasBranches(Mapping):
 
         return library.group(output, expression_context, how)
 
+    def num_entries_for(
+        self,
+        memory_size,
+        expressions=None,
+        cut=None,
+        filter_name=no_filter,
+        filter_typename=no_filter,
+        filter_branch=no_filter,
+        aliases=None,
+        compute=uproot4.compute.python.ComputePython(),
+        entry_start=None,
+        entry_stop=None,
+    ):
+        target_num_bytes = uproot4._util.memory_size(memory_size)
+
+        entry_start, entry_stop = _regularize_entries_start_stop(
+            self.tree.num_entries, entry_start, entry_stop
+        )
+
+        keys = set(self.keys(recursive=True, full_paths=False))
+        aliases = _regularize_aliases(self, aliases)
+        arrays, expression_context, branchid_interpretation = _regularize_expressions(
+            self,
+            expressions,
+            cut,
+            filter_name,
+            filter_typename,
+            filter_branch,
+            keys,
+            aliases,
+            compute,
+            (lambda branchname, interpretation: None),
+        )
+
+        return _hasbranches_num_entries_for(
+            self, target_num_bytes, entry_start, entry_stop, branchid_interpretation
+        )
+
+    def iterate(
+        self,
+        expressions=None,
+        cut=None,
+        filter_name=no_filter,
+        filter_typename=no_filter,
+        filter_branch=no_filter,
+        aliases=None,
+        compute=uproot4.compute.python.ComputePython(),
+        entry_start=None,
+        entry_stop=None,
+        step_size="100 MB",
+        decompression_executor=None,
+        interpretation_executor=None,
+        library="ak",
+        how=None,
+        report=False,
+    ):
+        keys = set(self.keys(recursive=True, full_paths=False))
+        if isinstance(self, TBranch) and expressions is None and len(keys) == 0:
+            filter_branch = uproot4._util.regularize_filter(filter_branch)
+            for x in self.parent.iterate(
+                expressions=expressions,
+                cut=cut,
+                filter_name=filter_name,
+                filter_typename=filter_typename,
+                filter_branch=lambda branch: branch is self and filter_branch(branch),
+                aliases=aliases,
+                compute=compute,
+                entry_start=entry_start,
+                entry_stop=entry_stop,
+                step_size=step_size,
+                decompression_executor=decompression_executor,
+                interpretation_executor=interpretation_executor,
+                library=library,
+                how=how,
+                report=report,
+            ):
+                yield x
+
+        else:
+            entry_start, entry_stop = _regularize_entries_start_stop(
+                self.tree.num_entries, entry_start, entry_stop
+            )
+            decompression_executor, interpretation_executor = _regularize_executors(
+                decompression_executor, interpretation_executor
+            )
+            library = uproot4.interpretation.library._regularize_library(library)
+
+            aliases = _regularize_aliases(self, aliases)
+            (
+                arrays,
+                expression_context,
+                branchid_interpretation,
+            ) = _regularize_expressions(
+                self,
+                expressions,
+                cut,
+                filter_name,
+                filter_typename,
+                filter_branch,
+                keys,
+                aliases,
+                compute,
+                (lambda branchname, interpretation: None),
+            )
+
+            entry_step = _regularize_step_size(
+                self, step_size, entry_start, entry_stop, branchid_interpretation
+            )
+
+            if report:
+                tree = self.tree
+
+            previous_baskets = {}
+            for sub_entry_start in range(entry_start, entry_stop, entry_step):
+                sub_entry_stop = min(sub_entry_start + entry_step, entry_stop)
+                if sub_entry_stop - sub_entry_start == 0:
+                    continue
+
+                ranges_or_baskets = []
+                for expression, context in expression_context:
+                    branch = context.get("branch")
+                    if branch is not None and not context["is_duplicate"]:
+                        for (
+                            basket_num,
+                            range_or_basket,
+                        ) in branch.entries_to_ranges_or_baskets(
+                            sub_entry_start, sub_entry_stop
+                        ):
+                            previous_basket = previous_baskets.get(
+                                (id(branch), basket_num)
+                            )
+                            if previous_basket is None:
+                                ranges_or_baskets.append(
+                                    (branch, basket_num, range_or_basket)
+                                )
+                            else:
+                                ranges_or_baskets.append(
+                                    (branch, basket_num, previous_basket)
+                                )
+
+                arrays = {}
+                _ranges_or_baskets_to_arrays(
+                    self,
+                    ranges_or_baskets,
+                    branchid_interpretation,
+                    sub_entry_start,
+                    sub_entry_stop,
+                    decompression_executor,
+                    interpretation_executor,
+                    library,
+                    arrays,
+                )
+
+                output = compute.compute_expressions(
+                    arrays,
+                    expression_context,
+                    keys,
+                    aliases,
+                    self.file.file_path,
+                    self.object_path,
+                )
+
+                expression_context = [
+                    (e, c)
+                    for e, c in expression_context
+                    if c["is_primary"] and not c["is_cut"]
+                ]
+
+                arrays = library.group(output, expression_context, how)
+
+                if report:
+                    yield arrays, Report(
+                        sub_entry_start,
+                        sub_entry_stop,
+                        sub_entry_start,
+                        sub_entry_stop,
+                        self,
+                        tree,
+                        self.file,
+                        self.file.file_path,
+                    )
+                else:
+                    yield arrays
+
+                for branch, basket_num, basket in ranges_or_baskets:
+                    previous_baskets[id(branch), basket_num] = basket
+
 
 _branch_clean_name = re.compile(r"(.*\.)*([^\.\[\]]*)(\[.*\])*")
 _branch_clean_parent_name = re.compile(r"(.*\.)*([^\.\[\]]*)\.([^\.\[\]]*)(\[.*\])*")
@@ -903,8 +1174,6 @@ class TBranch(HasBranches):
 
         self._lookup = {}
         self._interpretation = None
-        self._count_branch = None
-        self._count_leaf = None
         self._typename = None
         self._streamer = None
         self._context = dict(context)
@@ -1086,15 +1355,18 @@ in file {3}""".format(
 
     @property
     def count_branch(self):
-        if self._count_branch is None:
-            raise NotImplementedError
-        return self._count_branch
+        leaf = self.count_leaf
+        if leaf is None:
+            return None
+        else:
+            return leaf.parent
 
     @property
     def count_leaf(self):
-        if self._count_leaf is None:
-            raise NotImplementedError
-        return self._count_leaf
+        leaves = self.member("fLeaves")
+        if len(leaves) != 1:
+            return None
+        return leaves[0].member("fLeafCount")
 
     @property
     def num_entries(self):
@@ -1106,26 +1378,21 @@ in file {3}""".format(
 
     def __repr__(self):
         if len(self) == 0:
-            return "<TBranch {0} at 0x{1:012x}>".format(repr(self.name), id(self))
+            return "<{0} {1} at 0x{2:012x}>".format(
+                self.classname, repr(self.name), id(self)
+            )
         else:
-            return "<TBranch {0} ({1} subbranches) at 0x{2:012x}>".format(
-                repr(self.name), len(self), id(self)
+            return "<{0} {1} ({2} subbranches) at 0x{3:012x}>".format(
+                self.classname, repr(self.name), len(self), id(self)
             )
 
-    def basket_chunk_bytes(self, basket_num):
+    def basket_compressed_bytes(self, basket_num):
         if 0 <= basket_num < self._num_normal_baskets:
             return int(self.member("fBasketBytes")[basket_num])
         elif 0 <= basket_num < self.num_baskets:
-            raise IndexError(
-                """branch {0} has {1} normal baskets; cannot get """
-                """basket chunk {2} because only normal baskets have chunks
-in file {3}""".format(
-                    repr(self.name),
-                    self._num_normal_baskets,
-                    basket_num,
-                    self._file.file_path,
-                )
-            )
+            return self.embedded_baskets[
+                basket_num - self._num_normal_baskets
+            ].compressed_bytes
         else:
             raise IndexError(
                 """branch {0} has {1} baskets; cannot get basket chunk {2}
@@ -1137,7 +1404,7 @@ in file {3}""".format(
     def basket_chunk_cursor(self, basket_num):
         if 0 <= basket_num < self._num_normal_baskets:
             start = self.member("fBasketSeek")[basket_num]
-            stop = start + self.basket_chunk_bytes(basket_num)
+            stop = start + self.basket_compressed_bytes(basket_num)
             cursor = uproot4.source.cursor.Cursor(start)
             chunk = self._file.source.chunk(start, stop)
             return chunk, cursor
@@ -1194,7 +1461,7 @@ in file {3}""".format(
             if entry_start < stop and start <= entry_stop:
                 if 0 <= basket_num < self._num_normal_baskets:
                     byte_start = self.member("fBasketSeek")[basket_num]
-                    byte_stop = byte_start + self.basket_chunk_bytes(basket_num)
+                    byte_stop = byte_start + self.basket_compressed_bytes(basket_num)
                     out.append((basket_num, (byte_start, byte_stop)))
                 elif 0 <= basket_num < self.num_baskets:
                     out.append((basket_num, self.basket(basket_num)))
@@ -1203,6 +1470,37 @@ in file {3}""".format(
             start = stop
         return out
 
+    def debug(
+        self,
+        entry,
+        skip_bytes=None,
+        limit_bytes=None,
+        dtype=None,
+        offset=0,
+        stream=sys.stdout,
+    ):
+        interpretation = uproot4.interpretation.jagged.AsJagged(
+            uproot4.interpretation.numerical.AsDtype("u1")
+        )
+        data = self.array(
+            interpretation, entry_start=entry, entry_stop=entry + 1, library="np"
+        )[0]
+        chunk = uproot4.source.chunk.Chunk.wrap(self._file.source, data)
+        if skip_bytes is None:
+            cursor = uproot4.source.cursor.Cursor(0)
+        else:
+            cursor = uproot4.source.cursor.Cursor(skip_bytes)
+        cursor.debug(
+            chunk, limit_bytes=limit_bytes, dtype=dtype, offset=offset, stream=stream
+        )
+
+    def __array__(self, *args, **kwargs):
+        out = self.array(library="np")
+        if args == () and kwargs == {}:
+            return out
+        else:
+            return numpy.array(out, *args, **kwargs)
+
     def array(
         self,
         interpretation=None,
@@ -1263,3 +1561,417 @@ in file {3}""".format(
             array_cache[cache_key] = arrays[id(self)]
 
         return arrays[id(self)]
+
+
+_regularize_files_braces = re.compile(r"{([^}]*,)*([^}]*)}")
+
+
+def _regularize_files(files):
+    files = uproot4._util.regularize_path(files)
+
+    if uproot4._util.isstr(files):
+        file_path, object_path = uproot4._util.file_object_path_split(files)
+        parsed_url = urlparse(file_path)
+        count = 0
+
+        if parsed_url.scheme.upper() in uproot4._util._remote_schemes:
+            yield file_path, object_path
+            count += 1
+
+        else:
+            expanded = os.path.expanduser(file_path)
+            matches = list(_regularize_files_braces.finditer(expanded))
+            if len(matches) == 0:
+                results = [expanded]
+            else:
+                results = []
+                for combination in itertools.product(
+                    *[match.group(0)[1:-1].split(",") for match in matches]
+                ):
+                    tmp = expanded
+                    for c, m in list(zip(combination, matches))[::-1]:
+                        tmp = tmp[: m.span()[0]] + c + tmp[m.span()[1] :]
+                    results.append(tmp)
+
+            seen = set()
+            for result in results:
+                for match in glob.glob(result):
+                    if match not in seen:
+                        yield match, object_path
+                        seen.add(match)
+                        count += 1
+
+        if count == 0:
+            if hasattr(__builtins__, "FileNotFoundError"):
+                errclass = __builtins__.FileNotFoundError
+            else:
+                errclass = __builtins__.IOError
+            raise errclass("{0} did not match any files".format(repr(file_path)))
+
+    elif isinstance(files, HasBranches):
+        yield files, None
+
+    elif isinstance(files, Iterable):
+        count = 0
+        seen = set()
+        for file in files:
+            for file_path, object_path in _regularize_files(file):
+                if uproot4._util.isstr(file_path):
+                    if file_path not in seen:
+                        yield file_path, object_path
+                        seen.add(file_path)
+                else:
+                    yield file_path, object_path
+                    count += 1
+
+        if count == 0:
+            if hasattr(__builtins__, "FileNotFoundError"):
+                errclass = __builtins__.FileNotFoundError
+            else:
+                errclass = __builtins__.IOError
+            raise errclass("at least one file path or URL must be provided")
+
+    else:
+        raise TypeError(
+            "'files' must be a file path/URL (string or Path) with a TTree/TBranch "
+            "object path (separated by a colon ':'), possibly with glob "
+            "patterns (for local files), TTree/TBranch objects, or an iterable "
+            "of such things, not {0}".format(repr(files))
+        )
+
+
+class _NoClose(object):
+    def __init__(self, hasbranches):
+        self.hasbranches = hasbranches
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exception_type, exception_value, traceback):
+        pass
+
+
+def iterate(
+    files,
+    expressions=None,
+    cut=None,
+    filter_name=no_filter,
+    filter_typename=no_filter,
+    filter_branch=no_filter,
+    aliases=None,
+    compute=uproot4.compute.python.ComputePython(),
+    step_size="100 MB",
+    decompression_executor=None,
+    interpretation_executor=None,
+    library="ak",
+    how=None,
+    report=False,
+    custom_classes=None,
+    **options
+):
+    files = list(_regularize_files(files))
+    if any(
+        uproot4._util.isstr(file_path) and object_path is None
+        for file_path, object_path in files
+    ):
+        raise TypeError(
+            "'files' must include a TTree/TBranch object path (separated by a "
+            "colon ':') to each glob pattern (if multiple are given)"
+        )
+
+    decompression_executor, interpretation_executor = _regularize_executors(
+        decompression_executor, interpretation_executor
+    )
+    library = uproot4.interpretation.library._regularize_library(library)
+
+    global_start = 0
+    for file_path, object_path in files:
+        if object_path is None:
+            hasbranches = _NoClose(file_path)
+        else:
+            file = uproot4.reading.ReadOnlyFile(
+                file_path,
+                object_cache=None,
+                array_cache=None,
+                custom_classes=custom_classes,
+                **options
+            )
+            try:
+                hasbranches = file.root_directory[object_path]
+            except KeyError:
+                continue
+
+        with hasbranches:
+            for item in hasbranches.iterate(
+                expressions=expressions,
+                cut=cut,
+                filter_name=filter_name,
+                filter_typename=filter_typename,
+                filter_branch=filter_branch,
+                aliases=aliases,
+                compute=compute,
+                step_size=step_size,
+                decompression_executor=decompression_executor,
+                interpretation_executor=interpretation_executor,
+                library=library,
+                how=how,
+                report=report,
+            ):
+                if report:
+                    arrays, local_report = item
+                    global_entry_start = local_report.tree_entry_start
+                    global_entry_stop = local_report.tree_entry_stop
+                    global_entry_start += global_start
+                    global_entry_stop += global_start
+                    global_report = type(local_report)(
+                        *((global_entry_start, global_entry_stop) + local_report[2:])
+                    )
+                    arrays = library.global_index(arrays, global_start)
+                    yield arrays, global_report
+
+                else:
+                    arrays = library.global_index(item, global_start)
+                    yield arrays
+
+            global_start += hasbranches.num_entries
+
+
+def concatenate(
+    files,
+    expressions=None,
+    cut=None,
+    filter_name=no_filter,
+    filter_typename=no_filter,
+    filter_branch=no_filter,
+    aliases=None,
+    compute=uproot4.compute.python.ComputePython(),
+    decompression_executor=None,
+    interpretation_executor=None,
+    array_cache=None,
+    library="ak",
+    how=None,
+    report=False,
+    custom_classes=None,
+    **options
+):
+    files = list(_regularize_files(files))
+    if any(
+        uproot4._util.isstr(file_path) and object_path is None
+        for file_path, object_path in files
+    ):
+        raise TypeError(
+            "'files' must include a TTree/TBranch object path (separated by a "
+            "colon ':') to each glob pattern (if multiple are given)"
+        )
+
+    decompression_executor, interpretation_executor = _regularize_executors(
+        decompression_executor, interpretation_executor
+    )
+    library = uproot4.interpretation.library._regularize_library(library)
+
+    all_arrays = []
+    global_start = 0
+    for file_path, object_path in files:
+        if object_path is None:
+            hasbranches = _NoClose(file_path)
+        else:
+            file = uproot4.reading.ReadOnlyFile(
+                file_path,
+                object_cache=None,
+                array_cache=None,
+                custom_classes=custom_classes,
+                **options
+            )
+            try:
+                hasbranches = file.root_directory[object_path]
+            except KeyError:
+                continue
+
+        with hasbranches:
+            arrays = hasbranches.arrays(
+                expressions=expressions,
+                cut=cut,
+                filter_name=filter_name,
+                filter_typename=filter_typename,
+                filter_branch=filter_branch,
+                aliases=aliases,
+                compute=compute,
+                decompression_executor=decompression_executor,
+                interpretation_executor=interpretation_executor,
+                array_cache=array_cache,
+                library=library,
+                how=how,
+            )
+            arrays = library.global_index(arrays, global_start)
+            all_arrays.append(arrays)
+            global_start += hasbranches.num_entries
+
+    return library.concatenate(all_arrays)
+
+
+def lazy(
+    files,
+    filter_name=no_filter,
+    filter_typename=no_filter,
+    filter_branch=no_filter,
+    recursive=True,
+    full_paths=False,
+    step_size="100 MB",
+    decompression_executor=None,
+    interpretation_executor=None,
+    array_cache="100 MB",
+    library="ak",
+    report=False,
+    custom_classes=None,
+    **options
+):
+    files = list(_regularize_files(files))
+    if any(
+        uproot4._util.isstr(file_path) and object_path is None
+        for file_path, object_path in files
+    ):
+        raise TypeError(
+            "'files' must include a TTree/TBranch object path (separated by a "
+            "colon ':') to each glob pattern (if multiple are given)"
+        )
+
+    decompression_executor, interpretation_executor = _regularize_executors(
+        decompression_executor, interpretation_executor
+    )
+    array_cache = _regularize_array_cache(array_cache, None)
+    library = uproot4.interpretation.library._regularize_library_lazy(library)
+    import awkward1
+
+    if array_cache is not None:
+        array_cache = awkward1.layout.ArrayCache(array_cache)
+
+    real_options = dict(options)
+    if "num_workers" not in real_options:
+        real_options["num_workers"] = 1
+    if "num_fallback_workers" not in real_options:
+        real_options["num_fallback_workers"] = 1
+
+    filter_branch = uproot4._util.regularize_filter(filter_branch)
+
+    hasbranches = []
+    common_keys = None
+    is_self = []
+
+    for file_path, object_path in files:
+        if object_path is None:
+            obj = file_path
+        else:
+            obj = uproot4.reading.open(
+                file_path,
+                object_cache=None,
+                array_cache=None,
+                custom_classes=custom_classes,
+                **real_options
+            )[object_path]
+
+        if isinstance(obj, TBranch) and len(obj.keys(recursive=True)) == 0:
+            original = obj
+            obj = obj.parent
+            is_self.append(True)
+
+            def real_filter_branch(branch):
+                return branch is original and filter_branch(branch)
+
+        else:
+            is_self.append(False)
+            real_filter_branch = filter_branch
+
+        hasbranches.append(obj)
+
+        new_keys = obj.keys(
+            recursive=recursive,
+            filter_name=filter_name,
+            filter_typename=filter_typename,
+            filter_branch=real_filter_branch,
+            full_paths=full_paths,
+        )
+
+        if common_keys is None:
+            common_keys = new_keys
+        else:
+            new_keys = set(new_keys)
+            common_keys = [key for key in common_keys if key in new_keys]
+
+    if len(common_keys) == 0 or not (all(is_self) or not any(is_self)):
+        raise ValueError(
+            "TTrees in\n\n    {0}\n\nhave no TBranches in common".format(
+                "\n    ".join(
+                    "{0}:{1}".format(
+                        f.file_path if o is None else f,
+                        f.object_path if o is None else o,
+                    )
+                    for f, o in files
+                )
+            )
+        )
+
+    partitions = []
+    global_offsets = [0]
+    global_cache_key = []
+    for obj in hasbranches:
+        entry_start, entry_stop = _regularize_entries_start_stop(
+            obj.tree.num_entries, None, None
+        )
+        branchid_interpretation = {}
+        for key in common_keys:
+            branch = obj[key]
+            branchid_interpretation[id(branch)] = branch.interpretation
+        entry_step = _regularize_step_size(
+            obj, step_size, entry_start, entry_stop, branchid_interpretation
+        )
+
+        for start in range(entry_start, entry_stop, entry_step):
+            stop = min(start + entry_step, entry_stop)
+            length = stop - start
+
+            fields = []
+            names = []
+            for key in common_keys:
+                branch = obj[key]
+                form = branchid_interpretation[id(branch)].awkward_form(
+                    obj.file, index_format="i64"
+                )
+                generator = awkward1.layout.ArrayGenerator(
+                    branch.array,
+                    (
+                        None,
+                        start,
+                        stop,
+                        decompression_executor,
+                        interpretation_executor,
+                        None,
+                        "ak",
+                    ),
+                    {},
+                    uproot4._util.awkward_form_remove_uproot(awkward1, form),
+                    length,
+                )
+                cache_key = "{0}:{1}:{2}-{3}:{4}".format(
+                    branch.cache_key,
+                    branchid_interpretation[id(branch)].cache_key,
+                    start,
+                    stop,
+                    library.name,
+                )
+                global_cache_key.append(cache_key)
+                virtualarray = awkward1.layout.VirtualArray(
+                    generator, cache=array_cache, cache_key=cache_key
+                )
+                fields.append(virtualarray)
+                names.append(key)
+
+            recordarray = awkward1.layout.RecordArray(fields, names, length)
+            partitions.append(recordarray)
+            global_offsets.append(global_offsets[-1] + length)
+
+    out = awkward1.partition.IrregularlyPartitionedArray(partitions, global_offsets[1:])
+    out = awkward1.Array(out)
+
+    return library.wrap_awkward_lazy(
+        out, common_keys, global_offsets, ",".join(global_cache_key)
+    )
