@@ -10,7 +10,8 @@ import numpy
 import uproot4.interpretation.jagged
 import uproot4.interpretation.strings
 import uproot4.interpretation.objects
-import uproot4.stl_containers
+import uproot4.containers
+import uproot4.extras
 
 
 class Library(object):
@@ -20,12 +21,17 @@ class Library(object):
        * `imported`: The imported library or raises a helpful "how to"
              message if it could not be imported.
        * `empty(shape, dtype)`: Make (possibly temporary) multi-basket storage.
-       * `finalize(array, branch, interpretation)`: Convert the internal
-             storage form into one appropriate for this library (NumPy array,
-             Pandas Series, etc.).
+       * `finalize(array, branch, interpretation, entry_start, entry_stop)`:
+             Convert the internal storage form into one appropriate for this
+             library (NumPy array, Pandas Series, etc.).
        * `group(arrays, expression_context, how)`: Combine arrays into a group,
              either a generic tuple or a grouping style appropriate for this
              library (NumPy array dict, Awkward RecordArray, etc.).
+       * `global_index(array, global_start)`: Add `global_start` to the array's
+             index (if any). The array is modified in-place (if possible) and
+             returned.
+       * `concatenate(all_arrays)`: Eagerly combine a list of arrays into one
+             array.
     """
 
     @property
@@ -35,7 +41,7 @@ class Library(object):
     def empty(self, shape, dtype):
         return numpy.empty(shape, dtype)
 
-    def finalize(self, array, branch, interpretation):
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
         raise AssertionError
 
     def group(self, arrays, expression_context, how):
@@ -51,6 +57,12 @@ class Library(object):
                 "dict)".format(self.name)
             )
 
+    def global_index(self, array, global_start):
+        return array
+
+    def concatenate(self, all_arrays):
+        raise AssertionError
+
     def __repr__(self):
         return repr(self.name)
 
@@ -67,7 +79,7 @@ class NumPy(Library):
 
         return numpy
 
-    def finalize(self, array, branch, interpretation):
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
         if isinstance(array, uproot4.interpretation.jagged.JaggedArray) and isinstance(
             array.content, uproot4.interpretation.objects.StridedObjectArray,
         ):
@@ -95,6 +107,31 @@ class NumPy(Library):
         else:
             return array
 
+    def concatenate(self, all_arrays):
+        if len(all_arrays) == 0:
+            return all_arrays
+
+        if isinstance(all_arrays[0], (tuple, list)):
+            keys = range(len(all_arrays[0]))
+        elif isinstance(all_arrays[0], dict):
+            keys = list(all_arrays[0])
+        else:
+            raise AssertionError(repr(all_arrays[0]))
+
+        to_concatenate = dict((k, []) for k in keys)
+        for arrays in all_arrays:
+            for k in keys:
+                to_concatenate[k].append(arrays[k])
+
+        concatenated = dict((k, numpy.concatenate(to_concatenate[k])) for k in keys)
+
+        if isinstance(all_arrays[0], tuple):
+            return tuple(concatenated[k] for k in keys)
+        elif isinstance(all_arrays[0], list):
+            return [concatenated[k] for k in keys]
+        elif isinstance(all_arrays[0], dict):
+            return concatenated
+
 
 def _strided_to_awkward(awkward1, path, interpretation, data):
     contents = []
@@ -107,7 +144,11 @@ def _strided_to_awkward(awkward1, path, interpretation, data):
             if isinstance(member, uproot4.interpretation.objects.AsStridedObjects):
                 contents.append(_strided_to_awkward(awkward1, p, member, data))
             else:
-                contents.append(awkward1.layout.NumpyArray(numpy.array(data[p])))
+                contents.append(
+                    awkward1.from_numpy(
+                        numpy.array(data[p]), regulararray=True, highlevel=False
+                    )
+                )
             names.append(name)
     parameters = {
         "__record__": uproot4.model.classname_decode(interpretation.model.__name__)[0]
@@ -217,18 +258,9 @@ class Awkward(Library):
 
     @property
     def imported(self):
-        try:
-            import awkward1
-        except ImportError:
-            raise ImportError(
-                """install the 'awkward1' package with:
+        return uproot4.extras.awkward1()
 
-    pip install awkward1"""
-            )
-        else:
-            return awkward1
-
-    def finalize(self, array, branch, interpretation):
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
         awkward1 = self.imported
 
         if isinstance(array, uproot4.interpretation.objects.StridedObjectArray):
@@ -253,7 +285,9 @@ class Awkward(Library):
             return awkward1.Array(layout)
 
         elif isinstance(array, uproot4.interpretation.jagged.JaggedArray):
-            content = awkward1.from_numpy(array.content, highlevel=False)
+            content = awkward1.from_numpy(
+                array.content, regulararray=True, highlevel=False
+            )
             if issubclass(array.offsets.dtype.type, numpy.int32):
                 offsets = awkward1.layout.Index32(array.offsets)
                 layout = awkward1.layout.ListOffsetArray32(offsets, content)
@@ -320,14 +354,18 @@ in object {3}""".format(
             array = array.reshape(-1)
             contents = []
             for name in array.dtype.names:
-                contents.append(awkward1.layout.NumpyArray(numpy.array(array[name])))
+                contents.append(
+                    awkward1.from_numpy(
+                        numpy.array(array[name]), regulararray=True, highlevel=False
+                    )
+                )
             out = awkward1.layout.RecordArray(contents, array.dtype.names, length)
             for size in shape[::-1]:
                 out = awkward1.layout.RegularArray(out, size)
             return awkward1.Array(out)
 
         else:
-            return awkward1.from_numpy(array)
+            return awkward1.from_numpy(array, regulararray=True)
 
     def group(self, arrays, expression_context, how):
         awkward1 = self.imported
@@ -372,7 +410,10 @@ in object {3}""".format(
                 cut = len(jagged[0])
                 for name in jagged:
                     cut = min(cut, len(name))
-                    while cut > 0 and name[:cut] != jagged[0][:cut]:
+                    while cut > 0 and (
+                        name[:cut] != jagged[0][:cut]
+                        or name[cut - 1] not in ("_", ".", "/")
+                    ):
                         cut -= 1
                     if cut == 0:
                         break
@@ -399,6 +440,37 @@ in object {3}""".format(
                 "None, for an unzipped record array".format(self.name)
             )
 
+    def concatenate(self, all_arrays):
+        awkward1 = self.imported
+
+        if len(all_arrays) == 0:
+            return all_arrays
+
+        if isinstance(all_arrays[0], (tuple, list)):
+            keys = range(len(all_arrays[0]))
+        elif isinstance(all_arrays[0], dict):
+            keys = list(all_arrays[0])
+        else:
+            return awkward1.concatenate(all_arrays)
+
+        to_concatenate = dict((k, []) for k in keys)
+        for arrays in all_arrays:
+            for k in keys:
+                to_concatenate[k].append(arrays[k])
+
+        concatenated = dict((k, awkward1.concatenate(to_concatenate[k])) for k in keys)
+
+        if isinstance(all_arrays[0], tuple):
+            return tuple(concatenated[k] for k in keys)
+        elif isinstance(all_arrays[0], list):
+            return [concatenated[k] for k in keys]
+        elif isinstance(all_arrays[0], dict):
+            return concatenated
+
+    def wrap_awkward_lazy(self, layout, common_keys, global_offsets, global_cache_key):
+        awkward1 = self.imported
+        return awkward1.Array(layout)
+
 
 def _pandas_rangeindex():
     import pandas
@@ -417,27 +489,21 @@ def _strided_to_pandas(path, interpretation, data, arrays, columns):
                 columns.append(p)
 
 
+def _pandas_basic_index(pandas, entry_start, entry_stop):
+    if hasattr(pandas, "RangeIndex"):
+        return pandas.RangeIndex(entry_start, entry_stop)
+    else:
+        return pandas.Int64Index(range(entry_start, entry_stop))
+
+
 class Pandas(Library):
     name = "pd"
 
     @property
     def imported(self):
-        try:
-            import pandas
-        except ImportError:
-            raise ImportError(
-                """install the 'pandas' package with:
+        return uproot4.extras.pandas()
 
-    pip install pandas
-
-or
-
-    conda install pandas"""
-            )
-        else:
-            return pandas
-
-    def finalize(self, array, branch, interpretation):
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
         pandas = self.imported
 
         if isinstance(array, uproot4.interpretation.objects.StridedObjectArray):
@@ -451,7 +517,10 @@ or
                 columns = pandas.MultiIndex.from_tuples(
                     [x + ("",) * (maxlen - len(x)) for x in columns]
                 )
-            out = pandas.DataFrame(dict(zip(columns, arrays)), columns=columns)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            out = pandas.DataFrame(
+                dict(zip(columns, arrays)), columns=columns, index=index
+            )
             out.leaflist = maxlen != 1
             return out
 
@@ -461,7 +530,8 @@ or
             array.content, uproot4.interpretation.objects.StridedObjectArray
         ):
             index = pandas.MultiIndex.from_arrays(
-                array.parents_localindex(), names=["entry", "subentry"]
+                array.parents_localindex(entry_start, entry_stop),
+                names=["entry", "subentry"],
             )
             arrays = []
             columns = []
@@ -483,7 +553,8 @@ or
 
         elif isinstance(array, uproot4.interpretation.jagged.JaggedArray):
             index = pandas.MultiIndex.from_arrays(
-                array.parents_localindex(), names=["entry", "subentry"]
+                array.parents_localindex(entry_start, entry_stop),
+                names=["entry", "subentry"],
             )
             return pandas.Series(array.content, index=index)
 
@@ -497,7 +568,8 @@ or
             out = numpy.zeros(len(array), dtype=numpy.object)
             for i, x in enumerate(array):
                 out[i] = x
-            return pandas.Series(out)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            return pandas.Series(out, index=index)
 
         elif array.dtype.names is not None and len(array.shape) != 1:
             names = []
@@ -507,14 +579,16 @@ or
                     name = (n + "".join("[" + str(x) + "]" for x in tup),)
                     names.append(name)
                     arrays[name] = array[n][(slice(None),) + tup]
-            out = pandas.DataFrame(arrays, columns=names)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            out = pandas.DataFrame(arrays, columns=names, index=index)
             out.leaflist = True
             return out
 
         elif array.dtype.names is not None:
             columns = pandas.MultiIndex.from_tuples([(x,) for x in array.dtype.names])
             arrays = dict((y, array[x]) for x, y in zip(array.dtype.names, columns))
-            out = pandas.DataFrame(arrays, columns=columns)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            out = pandas.DataFrame(arrays, columns=columns, index=index)
             out.leaflist = True
             return out
 
@@ -525,12 +599,14 @@ or
                 name = "".join("[" + str(x) + "]" for x in tup)
                 names.append(name)
                 arrays[name] = array[(slice(None),) + tup]
-            out = pandas.DataFrame(arrays, columns=names)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            out = pandas.DataFrame(arrays, columns=names, index=index)
             out.leaflist = False
             return out
 
         else:
-            return pandas.Series(array)
+            index = _pandas_basic_index(pandas, entry_start, entry_stop)
+            return pandas.Series(array, index=index)
 
     def _only_series(self, original_arrays, original_names):
         pandas = self.imported
@@ -658,32 +734,74 @@ or
                 "DataFrames without merging)".format(self.name)
             )
 
+    def global_index(self, arrays, global_start):
+        if type(arrays.index).__name__ == "MultiIndex":
+            if hasattr(arrays.index.levels[0], "arrays"):
+                index = arrays.index.levels[0].arrays  # pandas>=0.24.0
+            else:
+                index = arrays.index.levels[0].values  # pandas<0.24.0
+            numpy.add(index, global_start, out=index)
+
+        elif type(arrays.index).__name__ == "RangeIndex":
+            if hasattr(arrays.index, "start") and hasattr(arrays.index, "stop"):
+                index_start = arrays.index.start  # pandas>=0.25.0
+                index_stop = arrays.index.stop
+            else:
+                index_start = arrays.index._start  # pandas<0.25.0
+                index_stop = arrays.index._stop
+            arrays.index = type(arrays.index)(
+                index_start + global_start, index_stop + global_start
+            )
+
+        else:
+            if hasattr(arrays.index, "arrays"):
+                index = arrays.index.arrays  # pandas>=0.24.0
+            else:
+                index = arrays.index.values  # pandas<0.24.0
+            numpy.add(index, global_start, out=index)
+
+        return arrays
+
+    def concatenate(self, all_arrays):
+        pandas = self.imported
+
+        if len(all_arrays) == 0:
+            return all_arrays
+
+        if isinstance(all_arrays[0], (tuple, list)):
+            keys = range(len(all_arrays[0]))
+        elif isinstance(all_arrays[0], dict):
+            keys = list(all_arrays[0])
+        else:
+            return pandas.concat(all_arrays)
+
+        to_concatenate = dict((k, []) for k in keys)
+        for arrays in all_arrays:
+            for k in keys:
+                to_concatenate[k].append(arrays[k])
+
+        concatenated = dict((k, pandas.concat(to_concatenate[k])) for k in keys)
+
+        if isinstance(all_arrays[0], tuple):
+            return tuple(concatenated[k] for k in keys)
+        elif isinstance(all_arrays[0], list):
+            return [concatenated[k] for k in keys]
+        elif isinstance(all_arrays[0], dict):
+            return concatenated
+
 
 class CuPy(Library):
     name = "cp"
 
     @property
     def imported(self):
-        try:
-            import cupy
-        except ImportError:
-            raise ImportError(
-                """install the 'cupy' package with:
-
-    pip install cupy
-
-or
-
-    conda install cupy"""
-            )
-        else:
-            return cupy
+        return uproot4.extras.cupy()
 
     def empty(self, shape, dtype):
         cupy = self.imported
         return cupy.empty(shape, dtype)
 
-    def finalize(self, array, branch, interpretation):
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
         cupy = self.imported
 
         if isinstance(
@@ -700,6 +818,33 @@ or
             assert isinstance(array, cupy.ndarray)
             return array
 
+    def concatenate(self, all_arrays):
+        cupy = self.imported
+
+        if len(all_arrays) == 0:
+            return all_arrays
+
+        if isinstance(all_arrays[0], (tuple, list)):
+            keys = range(len(all_arrays[0]))
+        elif isinstance(all_arrays[0], dict):
+            keys = list(all_arrays[0])
+        else:
+            raise AssertionError(repr(all_arrays[0]))
+
+        to_concatenate = dict((k, []) for k in keys)
+        for arrays in all_arrays:
+            for k in keys:
+                to_concatenate[k].append(arrays[k])
+
+        concatenated = dict((k, cupy.concatenate(to_concatenate[k])) for k in keys)
+
+        if isinstance(all_arrays[0], tuple):
+            return tuple(concatenated[k] for k in keys)
+        elif isinstance(all_arrays[0], list):
+            return [concatenated[k] for k in keys]
+        elif isinstance(all_arrays[0], dict):
+            return concatenated
+
 
 _libraries = {
     NumPy.name: NumPy(),
@@ -732,13 +877,188 @@ _libraries["CUPY"] = _libraries[CuPy.name]
 
 def _regularize_library(library):
     if isinstance(library, Library):
-        return _libraries[library.name]
+        if library.name in _libraries:
+            return _libraries[library.name]
+        else:
+            raise ValueError(
+                "library {0} ({1}) cannot be used in this function".format(
+                    type(library).__name__, repr(library.name)
+                )
+            )
 
     elif isinstance(library, type) and issubclass(library, Library):
-        return _libraries[library().name]
+        if library().name in _libraries:
+            return _libraries[library().name]
+        else:
+            raise ValueError(
+                "library {0} ({1}) cannot be used in this function".format(
+                    library.__name__, repr(library().name)
+                )
+            )
 
     else:
         try:
             return _libraries[library]
         except KeyError:
-            raise ValueError("unrecognized library: {0}".format(repr(library)))
+            raise ValueError(
+                """library {0} not recognized (for this function); """
+                """try "np" (NumPy), "ak" (Awkward1), "pd" (Pandas), or "cp" (CuPy) """
+                """instead""".format(repr(library))
+            )
+
+
+_libraries_lazy = {Awkward.name: _libraries[Awkward.name]}
+
+
+class DaskArray(Library):
+    name = "da"
+
+    awkward = _libraries_lazy[Awkward.name]
+
+    @property
+    def imported(self):
+        return uproot4.extras.dask_array()
+
+    def empty(self, shape, dtype):
+        return self.awkward.empty(shape, dtype)
+
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
+        return self.awkward.finalize(
+            array, branch, interpretation, entry_start, entry_stop
+        )
+
+    def group(self, arrays, expression_context, how):
+        return self.awkward.group(arrays, expression_context, how)
+
+    def global_index(self, array, global_start):
+        return self.awkward.global_index(array, global_start)
+
+    def concatenate(self, all_arrays):
+        return self.awkward.concatenate(self, all_arrays)
+
+    def wrap_awkward_lazy(self, layout, common_keys, global_offsets, global_cache_key):
+        awkward1 = self.awkward.imported
+
+        if len(common_keys) == 1:
+            array = awkward1.Array(layout[common_keys[0]])
+        else:
+            array = awkward1.Array(layout)
+
+        dask_array = self.imported
+        return dask_array.from_array(
+            array,
+            chunks=[
+                global_offsets[i + 1] - global_offsets[i]
+                for i in range(len(global_offsets) - 1)
+            ],
+            name="ak-{0}".format(abs(hash(global_cache_key))),
+            asarray=False,
+            fancy=True,
+        )
+
+
+_libraries_lazy[DaskArray.name] = DaskArray()
+
+
+class DaskFrame(Library):
+    name = "dd"
+
+    awkward = _libraries_lazy[Awkward.name]
+    dask_array = _libraries_lazy[DaskArray.name]
+
+    @property
+    def imported(self):
+        return uproot4.extras.dask_dataframe()
+
+    def empty(self, shape, dtype):
+        return self.awkward.empty(shape, dtype)
+
+    def finalize(self, array, branch, interpretation, entry_start, entry_stop):
+        return self.awkward.finalize(
+            array, branch, interpretation, entry_start, entry_stop
+        )
+
+    def group(self, arrays, expression_context, how):
+        return self.awkward.group(arrays, expression_context, how)
+
+    def global_index(self, array, global_start):
+        return self.awkward.global_index(array, global_start)
+
+    def concatenate(self, all_arrays):
+        return self.awkward.concatenate(self, all_arrays)
+
+    def wrap_awkward_lazy(self, layout, common_keys, global_offsets, global_cache_key):
+        awkward1 = self.awkward.imported
+        dask_array = self.dask_array.imported
+        dask_dataframe = self.imported
+
+        series = []
+        for name in common_keys:
+            array = dask_array.from_array(
+                awkward1.Array(layout[name]),
+                chunks=[
+                    global_offsets[i + 1] - global_offsets[i]
+                    for i in range(len(global_offsets) - 1)
+                ],
+                name="ak-{0}".format(abs(hash((global_cache_key, name)))),
+                asarray=False,
+                fancy=True,
+            )
+            series.append(dask_dataframe.from_dask_array(array, columns=name))
+
+        return dask_dataframe.concat(series, axis=1)
+
+
+_libraries_lazy[DaskFrame.name] = DaskFrame()
+
+
+_libraries_lazy["awkward1"] = _libraries_lazy[Awkward.name]
+_libraries_lazy["Awkward1"] = _libraries_lazy[Awkward.name]
+_libraries_lazy["AWKWARD1"] = _libraries_lazy[Awkward.name]
+_libraries_lazy["awkward"] = _libraries_lazy[Awkward.name]
+_libraries_lazy["Awkward"] = _libraries_lazy[Awkward.name]
+_libraries_lazy["AWKWARD"] = _libraries_lazy[Awkward.name]
+
+_libraries_lazy["dask"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["Dask"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["DASK"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["dask.array"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["dask-array"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["dask_array"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["DaskArray"] = _libraries_lazy[DaskArray.name]
+_libraries_lazy["dask.dataframe"] = _libraries_lazy[DaskFrame.name]
+_libraries_lazy["dask-dataframe"] = _libraries_lazy[DaskFrame.name]
+_libraries_lazy["dask_dataframe"] = _libraries_lazy[DaskFrame.name]
+_libraries_lazy["DaskDataframe"] = _libraries_lazy[DaskFrame.name]
+
+
+def _regularize_library_lazy(library):
+    if isinstance(library, Library):
+        if library.name in _libraries_lazy:
+            return _libraries_lazy[library.name]
+        else:
+            raise ValueError(
+                "library {0} ({1}) cannot be used in this function".format(
+                    type(library).__name__, repr(library.name)
+                )
+            )
+
+    elif isinstance(library, type) and issubclass(library, Library):
+        if library().name in _libraries_lazy:
+            return _libraries_lazy[library().name]
+        else:
+            raise ValueError(
+                "library {0} ({1}) cannot be used in this function".format(
+                    library.__name__, repr(library().name)
+                )
+            )
+
+    else:
+        try:
+            return _libraries_lazy[library]
+        except KeyError:
+            raise ValueError(
+                """library {0} not recognized (for this function); """
+                """try "ak" (Awkward1), "da" (dask.array), or "dd" (dask.dataframe) """
+                """instead""".format(repr(library))
+            )
