@@ -11,7 +11,6 @@ import uproot4.const
 import uproot4._util
 import uproot4.interpretation.objects
 
-
 bootstrap_classnames = [
     "TStreamerInfo",
     "TStreamerElement",
@@ -46,6 +45,33 @@ def bootstrap_classes():
     return custom_classes
 
 
+def reset_classes():
+    if uproot4._util.py2:
+        reload = __builtins__["reload"]
+    else:
+        from importlib import reload
+
+    uproot4.classes = {}
+    uproot4.unknown_classes = {}
+
+    reload(uproot4.streamers)
+    reload(uproot4.models.TObject)
+    reload(uproot4.models.TString)
+    reload(uproot4.models.TArray)
+    reload(uproot4.models.TNamed)
+    reload(uproot4.models.TList)
+    reload(uproot4.models.THashList)
+    reload(uproot4.models.TObjArray)
+    reload(uproot4.models.TObjString)
+    reload(uproot4.models.TAtt)
+    reload(uproot4.models.TRef)
+    reload(uproot4.models.TTree)
+    reload(uproot4.models.TBranch)
+    reload(uproot4.models.TLeaf)
+    reload(uproot4.models.TBasket)
+    reload(uproot4.models.RNTuple)
+
+
 class Model(object):
     class_streamer = None
 
@@ -128,7 +154,7 @@ class Model(object):
         )
 
     @classmethod
-    def awkward_form(cls, file, header=False, tobject_header=True):
+    def awkward_form(cls, file, index_format="i64", header=False, tobject_header=True):
         raise uproot4.interpretation.objects.CannotBeAwkward(
             classname_decode(cls.__name__)[0]
         )
@@ -209,6 +235,15 @@ class Model(object):
     def bases(self):
         return self._bases
 
+    def base(self, *cls):
+        out = []
+        for x in getattr(self, "_bases", []):
+            if isinstance(x, cls):
+                out.append(x)
+            if isinstance(x, Model):
+                out.extend(x.base(*cls))
+        return out
+
     def has_member(self, name, bases=True, recursive_bases=True):
         if name in self._members:
             return True
@@ -484,11 +519,9 @@ class DispatchByVersion(object):
         )
 
     @classmethod
-    def awkward_form(cls, file, header=False, tobject_header=True):
+    def awkward_form(cls, file, index_format="i64", header=False, tobject_header=True):
         versioned_cls = file.class_named(classname_decode(cls.__name__)[0], "max")
-        return versioned_cls.awkward_form(
-            file, header=header, tobject_header=tobject_header
-        )
+        return versioned_cls.awkward_form(file, index_format, header, tobject_header)
 
 
 _classname_encode_pattern = re.compile(br"[^a-zA-Z0-9]+")
