@@ -34,9 +34,9 @@ class Model_TString(uproot4.model.Model, str):
         return str(self)
 
     @classmethod
-    def awkward_form(cls, file, header=False, tobject_header=True):
-        return uproot4.stl_containers.AsString(False, typename="TString").awkward_form(
-            file, header, tobject_header
+    def awkward_form(cls, file, index_format="i64", header=False, tobject_header=True):
+        return uproot4.containers.AsString(False, typename="TString").awkward_form(
+            file, index_format, header, tobject_header
         )
 
 
