@@ -222,20 +222,20 @@ class Model_TStreamerInfo(uproot4.model.Model):
         ]
         awkward_form = [
             "    @classmethod",
-            "    def awkward_form(cls, file, header=False, tobject_header=True):",
+            "    def awkward_form(cls, file, index_format='i64', header=False, tobject_header=True):",
             "        from awkward1.forms import NumpyForm, ListOffsetForm, "
             "RegularForm, RecordForm",
             "        contents = {}",
             "        if header:",
             "            contents['@num_bytes'] = "
-            "uproot4._util.awkward_form(numpy.dtype('u4'), file, header, tobject_header)",
+            "uproot4._util.awkward_form(numpy.dtype('u4'), file, index_format, header, tobject_header)",
             "            contents['@instance_version'] = "
-            "uproot4._util.awkward_form(numpy.dtype('u2'), file, header, tobject_header)",
+            "uproot4._util.awkward_form(numpy.dtype('u2'), file, index_format, header, tobject_header)",
         ]
         fields = []
         formats = []
         dtypes = []
-        stl_containers = []
+        containers = []
         base_names_versions = []
         member_names = []
         class_flags = {}
@@ -251,7 +251,7 @@ class Model_TStreamerInfo(uproot4.model.Model):
                 fields,
                 formats,
                 dtypes,
-                stl_containers,
+                containers,
                 base_names_versions,
                 member_names,
                 class_flags,
@@ -284,7 +284,7 @@ class Model_TStreamerInfo(uproot4.model.Model):
         for i, dt in enumerate(dtypes):
             class_data.append("    _dtype{0} = {1}".format(i, dt))
 
-        for i, stl in enumerate(stl_containers):
+        for i, stl in enumerate(containers):
             class_data.append("    _stl_container{0} = {1}".format(i, stl))
 
         class_data.append(
@@ -436,7 +436,7 @@ class Model_TStreamerArtificial(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -458,7 +458,7 @@ class Model_TStreamerArtificial(Model_TStreamerElement):
         )
 
         awkward_form.append(
-            "        raise uproot4.deserialization.CannotBeAwkward("
+            "        raise uproot4.interpretation.objects.CannotBeAwkward("
             "'not implemented: class members defined by {0} of type {1} in member "
             "{2} of class {3}')".format(
                 type(self).__name__, self.typename, self.name, streamerinfo.name
@@ -512,7 +512,7 @@ class Model_TStreamerBase(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -531,7 +531,7 @@ class Model_TStreamerBase(Model_TStreamerElement):
         )
         awkward_form.append(
             "        contents.update(file.class_named({0}, {1}).awkward_form(file, "
-            "header, tobject_header).contents)".format(
+            "index_format, header, tobject_header).contents)".format(
                 repr(self.name), self.base_version
             )
         )
@@ -570,7 +570,7 @@ class Model_TStreamerBasicPointer(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -603,8 +603,8 @@ class Model_TStreamerBasicPointer(Model_TStreamerElement):
 
         awkward_form.extend(
             [
-                "        contents[{0}] = ListOffsetForm('i32', "
-                "uproot4._util.awkward_form(cls._dtype{1}, file, header, "
+                "        contents[{0}] = ListOffsetForm(index_format, "
+                "uproot4._util.awkward_form(cls._dtype{1}, file, index_format, header, "
                 "tobject_header),".format(repr(self.name), len(dtypes)),
                 "            parameters={'uproot': {'as': 'TStreamerBasicPointer', "
                 "'count_name': " + repr(self.count_name) + "}}",
@@ -697,7 +697,7 @@ class Model_TStreamerBasicType(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -791,7 +791,7 @@ class Model_TStreamerBasicType(Model_TStreamerElement):
             else:
                 awkward_form.append(
                     "        contents[{0}] = uproot4._util.awkward_form({1}, "
-                    "file, header, tobject_header)".format(
+                    "file, index_format, header, tobject_header)".format(
                         repr(self.name), _ftype_to_dtype(self.fType)
                     )
                 )
@@ -799,7 +799,7 @@ class Model_TStreamerBasicType(Model_TStreamerElement):
         else:
             awkward_form.append(
                 "        contents[{0}] = RegularForm(uproot4._util.awkward_form({1}, "
-                "file, header, tobject_header), {2})".format(
+                "file, index_format, header, tobject_header), {2})".format(
                     repr(self.name), _ftype_to_dtype(self.fType), self.array_length
                 )
             )
@@ -845,7 +845,7 @@ class Model_TStreamerLoop(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -872,8 +872,12 @@ class Model_TStreamerLoop(Model_TStreamerElement):
         awkward_form.extend(
             [
                 "        tmp = file.class_named({0}, 'max').awkward_form(file, "
-                "header, tobject_header)".format(repr(self.typename.rstrip("*"))),
-                "        contents[" + repr(self.name) + "] = ListOffsetForm('i32', "
+                "index_format, header, tobject_header)".format(
+                    repr(self.typename.rstrip("*"))
+                ),
+                "        contents["
+                + repr(self.name)
+                + "] = ListOffsetForm(index_format, "
                 "tmp, parameters={'uproot': {'as': TStreamerLoop, 'count_name': "
                 + repr(self.count_name)
                 + "}})",
@@ -930,7 +934,7 @@ class Model_TStreamerSTL(Model_TStreamerElement):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -945,22 +949,24 @@ class Model_TStreamerSTL(Model_TStreamerElement):
         read_members.append(
             "        self._members[{0}] = self._stl_container{1}.read("
             "chunk, cursor, context, self._file, self)"
-            "".format(repr(self.name), len(stl_containers))
+            "".format(repr(self.name), len(containers))
         )
 
         strided_interpretation.append(
             "        members.append(({0}, cls._stl_container{1}."
             "strided_interpretation(file, header, tobject_header)))".format(
-                repr(self.name), len(stl_containers)
+                repr(self.name), len(containers)
             )
         )
 
         awkward_form.append(
             "        contents[{0}] = cls._stl_container{1}.awkward_form(file, "
-            "header, tobject_header)".format(repr(self.name), len(stl_containers))
+            "index_format, header, tobject_header)".format(
+                repr(self.name), len(containers)
+            )
         )
 
-        stl_containers.append(stl_container)
+        containers.append(stl_container)
         member_names.append(self.name)
 
 
@@ -990,7 +996,7 @@ class pointer_types(object):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -1010,7 +1016,7 @@ class pointer_types(object):
             )
             awkward_form.append(
                 "        contents[{0}] = file.class_named({1}, 'max').awkward_form(file, "
-                "header, tobject_header)".format(
+                "index_format, header, tobject_header)".format(
                     repr(self.name), repr(self.typename.rstrip("*"))
                 )
             )
@@ -1078,7 +1084,7 @@ class object_types(object):
         fields,
         formats,
         dtypes,
-        stl_containers,
+        containers,
         base_names_versions,
         member_names,
         class_flags,
@@ -1096,7 +1102,7 @@ class object_types(object):
         )
         awkward_form.append(
             "        contents[{0}] = file.class_named({1}, 'max').awkward_form(file, "
-            "header, tobject_header)".format(
+            "index_format, header, tobject_header)".format(
                 repr(self.name), repr(self.typename.rstrip("*"))
             )
         )
